//There are a number of things that "ordered set" can mean; this implements the one people usually want, which is basically an array with element uniquing and O(1) contains() and indexOf()
public struct MyOrderedSet<Element : Hashable> : ExpressibleByArrayLiteral, MutableCollection, RangeReplaceableCollection {
    //Array and Dictionary take care of all the bounds checking and copy-on-write stuff for us. Nice.
    private var _indices:[Element : Int] = [:] //map values to the indices that they're stored at in the _values array
    private var _values:[Element] = [] //the actual values
    
    public var count : Int {
        return _values.count
    }
    
    public var startIndex : Int {
        return _values.startIndex //always going to be 0, but for symmetry's sake, may as well say it this way
    }
    
    public var endIndex : Int {
        return _values.endIndex
    }
    
    public func index(after i: Int) -> Int {
        return _values.index(after: i)
    }
    
    public func index(before i: Int) -> Int {
        return _values.index(before: i)
    }
    
    public init() {}
    
    public init<Seq:Sequence>(elements: Seq) where Seq.Iterator.Element == Element {
        replaceSubrange(self.startIndex..<self.endIndex, with: elements)
    }
    
    public init(arrayLiteral elements: Element...) {
        self.init(elements:elements)
    }
    
    public func index(of value:Element) -> Int? { //override the protocol extension version for performance, since we can do better
        return _indices[value]
    }
    
    public func contains(element: Element) -> Bool { //override the protocol extension version for performance, since we can do better
        return _indices[element] != nil
    }
    
    private mutating func remove(elementsIn subRange: Range<Int>) {
        for value in _values[subRange] {
            _indices[value] = nil
        }
        _values.replaceSubrange(subRange, with: EmptyCollection())
    }
    
    //We provide an overload of insert(contentsOf:at:) that takes a Sequence instead of a Collection so that we can use it inside of replaceSubrange
    public mutating func insert<Seq:Sequence>(contentsOf newElements: Seq, at idx:Int) where Seq.Iterator.Element == Element {
        var idx = idx
        for newValue in newElements {
            //I've decided to say that the first occurrence of a value that's added "wins". Could do the reverse by removing the element if already present, but that would mean recalculating the current index, and I'm lazy
            if index(of: newValue) == nil {
                _indices[newValue] = idx
                _values.insert(newValue, at: idx)
                idx += 1
            }
        }
    }
    
    //When we add or remove elements, we have to adjust the element->index mapping for the new indices of all the elements that moved
    private mutating func shiftIndices(startingAtIndex idx:Int, by distance:Int) {
        for value in _values.suffix(from: idx) {
            _indices[value]! += distance
        }
    }
    
    public mutating func replaceSubrange<Seq:Sequence>(_ subRange: Range<Int>, with newElements: Seq) where Seq.Iterator.Element == Element {
        //Step one: remove everything in the range being replaced
        remove(elementsIn: subRange)
        
        let oldCount = count
        //Step two: add the new elements at the start of the range
        insert(contentsOf: newElements, at: subRange.lowerBound)
        let addedElementsCount = count - oldCount
       
        //Step three: adjust our dictionary of elements to indices to reflect the new positions of things
        shiftIndices(startingAtIndex: subRange.lowerBound + addedElementsCount, by: addedElementsCount - subRange.count)
    }
    
    public subscript(index: Int) -> Element {
        get {
            return _values[index]
        }
        set {
            replaceSubrange(index...index, with: CollectionOfOne(newValue))
        }
    }
    
    //We provide an overload of remove() that takes an Element instead of an Index as a convenience, since it's a fast operation for ordered sets
    public mutating func remove(element:Element) -> Element? {
        if let index = index(of: element) {
            replaceSubrange(index...index, with: EmptyCollection())
            return element
        }
        return nil
    }
}

var testSet:MyOrderedSet<String> = ["weasel", "frog", "magnet", "ectoplasm", "frog", "weasel"]
print(testSet)
print(Array(testSet))
print(Array(testSet[1...2]))
testSet.append("glow worm")
var uppercaseSet = MyOrderedSet(elements: testSet.map { $0.uppercased() })
print(Array(uppercaseSet))
uppercaseSet.remove(element: "WEASEL")
uppercaseSet.remove(element: "FROG")
print(Array(uppercaseSet))
uppercaseSet[1] = "WOMBAT"
print(Array(uppercaseSet))
print(uppercaseSet.sorted())

